@@ -1,5 +1,8 @@
+import contextlib
 from typing import Dict, Any
 
+from contextvars import ContextVar
+
 
 class ContextData(object):
     def __init__(self, props: Dict[str, Any] = None, **kwargs):
@@ -12,7 +15,24 @@ class ContextData(object):
         if isinstance(value, (tuple, list, set, frozenset)):
             return type(value)([self._wrap(v) for v in value])
         else:
-            return ContextData(value) if isinstance(value, dict) else value
+            return value
 
     def __setattr__(self, key, value):
         super().__setattr__(key, self._wrap(value))
+
+
+_ctx: ContextVar[ContextData] = ContextVar('_ctx')
+
+
+def get_context():
+    if ctx := _ctx.get(None):
+        return ctx
+    _ctx.set(ContextData())
+    return _ctx.get()
+
+
+@contextlib.contextmanager
+def with_context():
+    token = _ctx.set(ContextData())
+    yield
+    _ctx.reset(token)
