@@ -1,46 +1,68 @@
+from dataclasses import dataclass
 from datetime import datetime
 from random import randint
 
 import pytest
+from dropland.blocks.redis import USE_REDIS
+from dropland.blocks.sql import USE_SQL
 from sqlalchemy import Column, DateTime, Integer, String, func, text
-
-from dropland.storages.redis import USE_REDIS
-from dropland.storages.sql import USE_SQL
+from tests.conftest import sql_storage
 
 pytestmark = pytest.mark.skipif(not USE_SQL, reason='For Sql only')
 
-if USE_REDIS:
-    from dropland.data.model import CachedModel
-    from .redis_models_data import redis_engine
 
 if USE_SQL:
-    from .sql_models_data import USE_GINO, USE_SQLA, \
-        gino_mysql, gino_pg, sqla_mysql, sqla_pg, sqla_sqlite
+    from dropland.blocks.sql import SqlStorageEngine, SqlaModel
+    from dropland.data.models.sql import CachedSqlModel
+    from .sql_models_data import sqla_mysql, sqla_pg, sqla_sqlite
+    if USE_REDIS:
+        from dropland.blocks.redis.model import RedisCacheType, RedisMethodCache
 
-    sqla_engines = (sqla_sqlite, sqla_pg, sqla_mysql)
-    gino_engines = (gino_pg, gino_mysql)
+    @dataclass
+    class ModelArgs:
+        table_name: str
+        sql_engine: SqlStorageEngine
+        model_class: type
 
-    engines = sqla_engines if USE_SQLA else gino_engines
-    models = [e.Model for e in engines]
-    ids = ['SQLite', 'PostgreSQL', 'MySQL'] if USE_SQLA else ['PostgreSQL', 'MySQL']
+    ids = ['SQLite', 'PostgreSQL', 'MySQL']
+    engines = [sqla_sqlite, sqla_pg, sqla_mysql]
+    models_args = []
+
+    for name, engine in zip(ids, engines):
+        models_args.append(
+            ModelArgs(table_name=f'testmodel_{name.lower()}', sql_engine=engine, model_class=SqlaModel)
+        )
 
     if USE_REDIS:
-        models += [CachedModel('dropland', e.Model, redis_engine.Model) for e in engines]
-        ids += [f'{i}+Redis' for i in ids]
+        from .redis_models_data import redis_engine
+
+        for name, engine in zip(ids, engines):
+            table_name = f'testmodel_{name.lower()}_redis_simple'
+
+            class CachedSqlaModel(CachedSqlModel, SqlaModel):
+                __abstract__ = True
+
+                class Meta(CachedSqlModel.Meta, SqlaModel.Meta):
+                    caches = [RedisMethodCache(redis_engine, table_name, RedisCacheType.SIMPLE)]
+
+            models_args.append(
+                ModelArgs(table_name=table_name, sql_engine=engine, model_class=CachedSqlaModel)
+            )
+
+        ids += [f'{i}+Redis(simple)' for i in ids]
 else:
-    models, ids = [], []
+    models_args, ids = [], []
 
 
 # noinspection PyPep8Naming
 @pytest.mark.asyncio
-@pytest.fixture(scope='module', params=models, ids=ids)
+@pytest.fixture(scope='module', params=models_args, ids=ids)
 async def TestModel(request):
-    Model = request.param
-    engine = request.param.get_engine()
-    table_name = f'testmodel_{ids[request.param_index].lower()}'
+    model_args: ModelArgs = request.param
+    engine = model_args.sql_engine
 
-    class SomeModel(Model):
-        __tablename__ = table_name
+    class SomeModel(model_args.model_class, sql_engine=engine):
+        __tablename__ = model_args.table_name
 
         id = Column(Integer, primary_key=True)
         data = Column(String(128))
@@ -53,36 +75,34 @@ async def TestModel(request):
 
     assert SomeModel.get_engine() is engine
     ti = SomeModel.get_table_info()
-    assert ti.backend == engine.backend.name
     assert ti.db_type == engine.db_type
     assert ti.is_async == engine.is_async
 
-    if USE_SQLA:
-        async with engine.raw_engine.begin() as conn:
-            await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
-            await conn.run_sync(engine.metadata.create_all, tables=[SomeModel.__table__])
-            await conn.commit()
+    async with engine.raw_engine.begin() as conn:
+        await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
+        await conn.run_sync(engine.metadata.create_all, tables=[SomeModel.__table__])
+        await conn.commit()
 
-        yield SomeModel
+    yield SomeModel
 
-        async with engine.raw_engine.begin() as conn:
-            await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
-            await conn.commit()
+    async with engine.raw_engine.begin() as conn:
+        await conn.run_sync(engine.metadata.drop_all, tables=[SomeModel.__table__])
+        await conn.commit()
 
-    elif USE_GINO:
-        from dropland.storages.sql import default_storage
-        from gino.schema import GinoSchemaVisitor
 
-        async with default_storage.manager().with_async_engines(['gino']):
-            bind = engine.raw_engine.bind
-            await GinoSchemaVisitor(engine.metadata).drop_all(bind, tables=[SomeModel.__table__])
-            await GinoSchemaVisitor(engine.metadata).create_all(bind, tables=[SomeModel.__table__])
+@pytest.fixture(scope='module')
+def wire_sql_storage():
+    sql_storage.wire()
+    yield
+    sql_storage.unwire()
 
-        yield SomeModel
 
-        async with default_storage.manager().with_async_engines(['gino']):
-            bind = engine.raw_engine.bind
-            await GinoSchemaVisitor(engine.metadata).drop_all(bind, tables=[SomeModel.__table__])
+@pytest.mark.asyncio
+@pytest.fixture(scope='function')
+async def model_session(event_loop, wire_sql_storage, redis_session):
+    session = await sql_storage.async_session_context(autocommit=False)
+    yield session
+    await sql_storage.async_session_context.shutdown()
 
 
 _construct_counter = 0
@@ -93,8 +113,7 @@ _build_rela_list_counter = 0
 
 @pytest.fixture(scope='function')
 def build_calls_counter(TestModel, monkeypatch):
-    global _construct_counter, _construct_list_counter, \
-        _build_rela_counter, _build_rela_list_counter
+    global _construct_counter, _construct_list_counter, _build_rela_counter, _build_rela_list_counter
     _construct_counter = 0
     _construct_list_counter = 0
     _build_rela_counter = 0
@@ -105,15 +124,15 @@ def build_calls_counter(TestModel, monkeypatch):
     old_build_rela = TestModel._build_rela
     old_build_rela_list = TestModel._build_rela_list
 
-    async def construct(*args, **kwargs):
+    def construct(*args, **kwargs):
         global _construct_counter
         _construct_counter += 1
-        return await old_construct(*args, **kwargs)
+        return old_construct(*args, **kwargs)
 
-    async def construct_list(*args, **kwargs):
+    def construct_list(*args, **kwargs):
         global _construct_list_counter
         _construct_list_counter += 1
-        return await old_construct_list(*args, **kwargs)
+        return old_construct_list(*args, **kwargs)
 
     async def build_rela(*args, **kwargs):
         global _build_rela_counter
@@ -132,7 +151,7 @@ def build_calls_counter(TestModel, monkeypatch):
 
 
 @pytest.mark.asyncio
-async def test_model_fields(data_ctx, TestModel):
+async def test_model_fields(TestModel, model_session):
     m = TestModel()
     assert TestModel.get_fields() == {'id', 'create_date', 'data'}
     assert m.get_values() == {'id': None, 'create_date': None, 'data': None}
@@ -143,19 +162,17 @@ async def test_model_fields(data_ctx, TestModel):
 
 
 @pytest.mark.asyncio
-async def test_create(data_ctx, TestModel, build_calls_counter):
+async def test_create(TestModel, build_calls_counter, model_session):
     obj = await TestModel.create(dict(data='123'))
     assert obj.id > 0
     assert obj.data == '123'
 
-    assert _construct_counter == 1
-    assert _build_rela_counter == _construct_counter
-    assert _construct_list_counter == 0
-    assert _build_rela_list_counter == _construct_list_counter
+    assert _build_rela_counter == 1
+    assert _build_rela_list_counter == 0
 
 
 @pytest.mark.asyncio
-async def test_get_or_create(data_ctx, TestModel, build_calls_counter):
+async def test_get_or_create(TestModel, build_calls_counter, model_session):
     obj1, created = await TestModel.get_or_create(999, dict(id=999, data='123'))
     assert created is True
     assert obj1.id == 999
@@ -166,44 +183,38 @@ async def test_get_or_create(data_ctx, TestModel, build_calls_counter):
     assert obj2.id == 999
     assert obj2.data == '123'
 
-    assert _construct_counter == 1
+    await TestModel.get(999)
+
+    assert _construct_counter == (3 if TestModel.has_cache() else 1)
     assert _build_rela_counter == _construct_counter
     assert _construct_list_counter == 0
-    assert _build_rela_list_counter == _construct_list_counter
+    assert _build_rela_list_counter == 0
 
 
 @pytest.mark.asyncio
-async def test_update(data_ctx, TestModel, build_calls_counter):
+async def test_update(TestModel, build_calls_counter, model_session):
     obj = await TestModel.create(dict(data='123'))
     assert await obj.update(dict(data='456'))
     assert obj.id > 0
     assert obj.data == '456'
 
-    assert _construct_counter == 2
-    assert _build_rela_counter == _construct_counter
-    assert _construct_list_counter == 0
-    assert _build_rela_list_counter == _construct_list_counter
+    assert _build_rela_counter == 1
+    assert _build_rela_list_counter == 0
 
 
 @pytest.mark.asyncio
-async def test_update_by_id(data_ctx, TestModel, build_calls_counter):
+async def test_update_by_id(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='123'))
 
-    obj2 = await TestModel.update_by_id(obj1.id, dict(data='234'))
-    assert obj2.id > 0
-    assert obj2.data == '234'
+    assert await TestModel.update_by_id(obj1.id, dict(data='234'))
+    assert not await TestModel.update_by_id(-1, dict(data='234'))
 
-    obj3 = await TestModel.update_by_id(-1, dict(data='234'))
-    assert obj3 is None
-
-    assert _construct_counter == 2
-    assert _build_rela_counter == _construct_counter
-    assert _construct_list_counter == 0
-    assert _build_rela_list_counter == _construct_list_counter
+    assert _build_rela_counter == 1
+    assert _build_rela_list_counter == 0
 
 
 @pytest.mark.asyncio
-async def test_update_by(data_ctx, TestModel, build_calls_counter):
+async def test_update_by(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -215,14 +226,12 @@ async def test_update_by(data_ctx, TestModel, build_calls_counter):
     assert obj2.data == '0'
     assert obj3.data == '0'
 
-    assert _construct_counter == 3
-    assert _build_rela_counter == _construct_counter
-    assert _construct_list_counter == 0
-    assert _build_rela_list_counter == _construct_list_counter
+    assert _build_rela_counter == 3
+    assert _build_rela_list_counter == 0
 
 
 @pytest.mark.asyncio
-async def test_get(data_ctx, TestModel, build_calls_counter):
+async def test_get(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='123'))
 
     obj2 = await TestModel.get(obj1.id)
@@ -233,14 +242,14 @@ async def test_get(data_ctx, TestModel, build_calls_counter):
     obj3 = await TestModel.get(-1)
     assert obj3 is None
 
-    assert _construct_counter == 1
+    assert _construct_counter == (2 if TestModel.has_cache() else 1)
     assert _build_rela_counter == _construct_counter
     assert _construct_list_counter == 0
-    assert _build_rela_list_counter == _construct_list_counter
+    assert _build_rela_list_counter == 0
 
 
 @pytest.mark.asyncio
-async def test_list(data_ctx, TestModel, build_calls_counter):
+async def test_list(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -301,7 +310,7 @@ async def test_list(data_ctx, TestModel, build_calls_counter):
 
 
 @pytest.mark.asyncio
-async def test_filter(data_ctx, TestModel, build_calls_counter):
+async def test_filter(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -349,7 +358,7 @@ async def test_filter(data_ctx, TestModel, build_calls_counter):
 
 
 @pytest.mark.asyncio
-async def test_any(data_ctx, TestModel, build_calls_counter):
+async def test_any(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -376,14 +385,16 @@ async def test_any(data_ctx, TestModel, build_calls_counter):
     assert objects[3] == obj3
     assert objects[4] is None
 
-    assert _construct_counter == 3
+    await TestModel.get_any([obj1.id, obj3.id])
+
+    assert _construct_counter == (12 if TestModel.has_cache() else 3)
     assert _build_rela_counter == 3
-    assert _construct_list_counter == (2 if TestModel.has_cache() else 1)
-    assert _build_rela_list_counter == 0
+    assert _construct_list_counter == (4 if TestModel.has_cache() else 1)
+    assert _build_rela_list_counter == (4 if TestModel.has_cache() else 0)
 
 
 @pytest.mark.asyncio
-async def test_count(data_ctx, TestModel, build_calls_counter):
+async def test_count(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -419,7 +430,7 @@ async def test_count(data_ctx, TestModel, build_calls_counter):
 
 
 @pytest.mark.asyncio
-async def test_exists(data_ctx, TestModel, build_calls_counter):
+async def test_exists(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -461,7 +472,7 @@ async def test_exists(data_ctx, TestModel, build_calls_counter):
 
 
 @pytest.mark.asyncio
-async def test_delete(data_ctx, TestModel, build_calls_counter):
+async def test_delete(TestModel, build_calls_counter, model_session):
     obj = await TestModel.create(dict(data='123'))
 
     objects = await TestModel.list()
@@ -479,14 +490,14 @@ async def test_delete(data_ctx, TestModel, build_calls_counter):
 
     assert (await obj.delete()) is False
 
-    assert _construct_counter == 2
-    assert _build_rela_counter == 1
+    assert _construct_counter == (3 if TestModel.has_cache() else 2)
+    assert _build_rela_counter == (2 if TestModel.has_cache() else 1)
     assert _construct_list_counter == 2
     assert _build_rela_list_counter == 1
 
 
 @pytest.mark.asyncio
-async def test_delete_by_id(data_ctx, TestModel, build_calls_counter):
+async def test_delete_by_id(TestModel, build_calls_counter, model_session):
     obj = await TestModel.create(dict(data='123'))
 
     objects = await TestModel.list()
@@ -505,14 +516,14 @@ async def test_delete_by_id(data_ctx, TestModel, build_calls_counter):
     assert (await TestModel.delete_by_id(obj.id)) is False
     assert (await TestModel.delete_by_id(-1)) is False
 
-    assert _construct_counter == 2
-    assert _build_rela_counter == 1
+    assert _construct_counter == (3 if TestModel.has_cache() else 2)
+    assert _build_rela_counter == (2 if TestModel.has_cache() else 1)
     assert _construct_list_counter == 2
     assert _build_rela_list_counter == 1
 
 
 @pytest.mark.asyncio
-async def test_delete_by(data_ctx, TestModel, build_calls_counter):
+async def test_delete_by(TestModel, build_calls_counter, model_session):
     obj1 = await TestModel.create(dict(data='1'))
     obj2 = await TestModel.create(dict(data='2'))
     obj3 = await TestModel.create(dict(data='3'))
@@ -527,14 +538,14 @@ async def test_delete_by(data_ctx, TestModel, build_calls_counter):
     assert not await obj2.load()
     assert not await obj3.load()
 
-    assert _construct_counter == 4
+    assert _construct_counter == 3
     assert _build_rela_counter == _construct_counter
     assert _construct_list_counter == 0
     assert _build_rela_list_counter == _construct_list_counter
 
 
 @pytest.mark.asyncio
-async def test_save_and_load(data_ctx, TestModel, build_calls_counter):
+async def test_save_and_load(TestModel, build_calls_counter, model_session):
     m = TestModel()
     m.id = randint(1, 1000000)
     m.data = 'qwer'
@@ -558,21 +569,21 @@ async def test_save_and_load(data_ctx, TestModel, build_calls_counter):
 
     m3 = await TestModel.get(1234)
     assert m3.id == 1234
-    assert m3.data == 'qwer'
+    assert m3.data == 'qwerty'
     assert m3.create_date.replace(microsecond=0) == m2.create_date.replace(microsecond=0)
 
     assert await m2.load()
     assert m2.id == 1234
-    assert m2.data == 'qwer'
+    assert m2.data == 'qwerty'
 
-    assert _construct_counter == 1
-    assert _build_rela_counter == 3
+    assert _construct_counter == 2
+    assert _build_rela_counter == _construct_counter
     assert _construct_list_counter == 0
     assert _build_rela_list_counter == _construct_list_counter
 
 
 @pytest.mark.asyncio
-async def test_save_all(data_ctx, TestModel, build_calls_counter):
+async def test_save_all(TestModel, build_calls_counter, model_session):
     obj1 = TestModel(data='1')
     obj2 = TestModel(data='2')
     obj3 = TestModel(data='3')
@@ -609,7 +620,7 @@ async def test_save_all(data_ctx, TestModel, build_calls_counter):
     assert objects[2].id == obj3.id
     assert objects[2].data == obj3.data
 
-    assert _construct_counter == 0
+    assert _construct_counter == (6 if TestModel.has_cache() else 0)
     assert _build_rela_counter == 0
-    assert _construct_list_counter == 0
+    assert _construct_list_counter == (2 if TestModel.has_cache() else 0)
     assert _build_rela_list_counter == _construct_list_counter
